#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <psapi.h>
#include <vector>
#include <string>
#include <util.h>

typedef std::vector< DWORD > pid_list_t;
enum version_t { XP, VISTA };

//
// Get a list of pids of the process running on the system.
//
pid_list_t get_process_pids()
{
    const unsigned BUF_SIZE = 4096;
    DWORD buf[BUF_SIZE];
    DWORD bytes_returned;
    if( !EnumProcesses( buf, sizeof(buf), &bytes_returned ) )
	return pid_list_t(); // TODO: increase buffer

    unsigned pids_returned = bytes_returned / sizeof(DWORD);
    return pid_list_t( buf, buf + pids_returned );
}

//
// Open the target process.
//
// name contains the name of the process (the exe), pids is the list
// of pids in the system. The function iterates through all the pids
// until it finds the process with the correct executable name.
//
// It is probably better to use CreateToolhelp32Snapshot() instead
// of doing it this way.
//
HANDLE open_target_process( const std::string& name, const pid_list_t& pids )
{
    for( pid_list_t::const_iterator iter = pids.begin();
	 iter != pids.end();
	 ++iter )
    {
	DWORD pid = *iter;
	HANDLE process = OpenProcess( PROCESS_QUERY_INFORMATION | PROCESS_VM_READ | PROCESS_VM_WRITE | PROCESS_CREATE_THREAD | PROCESS_VM_OPERATION, FALSE, pid );
	if( !process )
	    continue;

	// get first module
	HMODULE mod;
	DWORD bytes_needed;
	if( EnumProcessModules( process, &mod, sizeof(mod), &bytes_needed ) )
	{
	    char process_name[MAX_PATH];
	    
	    GetModuleBaseName( process, mod, process_name, sizeof(process_name) );
	    if( name == process_name )
		return process;
	}
	CloseHandle(process);
    }
    return 0;
}

//
// Injects a dll in the target process.
//
// This functions injects a dll in the target process. This is done by
// creating a thread in the process with the LoadLibrary as the thread
// function. Since LoadLibrary and the standard Win32 thread routing
// both takes one parameter, it is easy to send the name of the dll to
// load to LoadLibrary. 
// 
bool inject_dll( HANDLE process, const std::string& dll )
{
    // allocate memory for the dll name
    LPVOID addr = VirtualAllocEx( process, NULL, 4096, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE );
    if( !addr )
	return false;

    // copy dll name
    char buf[4096];
    memset( buf, 0, sizeof(buf) );
    memcpy( buf, dll.c_str(), dll.size() );
    if( !WriteProcessMemory( process, addr, buf, sizeof(buf), NULL ) )
	return false;

    // find address of LoadLibraryA
    // (the address is the same in all processes)
    FARPROC func = GetProcAddress( LoadLibrary("kernel32.dll"), "LoadLibraryA");
    if( !func )
	return false;

    // create thread
    DWORD thread_id;
    HANDLE thread = CreateRemoteThread( process, NULL, 0, (LPTHREAD_START_ROUTINE)func, addr, 0, &thread_id );
    if( !thread )
	return false;

    // wait for thread to finish
    WaitForSingleObject( thread, 10000 );
    return true;
}

std::string get_full_path( const std::string& str )
{
    char buf[MAX_PATH];
    GetFullPathName(str.c_str(), sizeof(buf), buf, NULL );
    return buf;
}

//
// Return the name of the solitaire executable for different windows
// versions.
//
const char* get_solitaire_name( version_t windows_version )
{
    if( windows_version == XP )
	return "sol.exe";
    else if( windows_version == VISTA )
	return "Solitaire.exe";
    return 0;
}

//
// Return the name of the trainer dll for different windows versions.
//
const char* get_dll_name( version_t windows_version )
{
    if( windows_version == XP )
	return "v!d-solxp_trainer.dll";
    else if( windows_version == VISTA )
	return "v!d-solvista_trainer.dll";
    return 0;
}

int main(int argc, char** argv)
{
    pid_list_t pids = get_process_pids();

    // get the windows version
    OSVERSIONINFO version_info;
    version_info.dwOSVersionInfoSize = sizeof(version_info);
    if( !GetVersionEx( &version_info ) )
	return 1;
    version_t windows_version;
    if( version_info.dwMajorVersion == 6 && version_info.dwMinorVersion == 0 )
	windows_version = VISTA;
    else if( version_info.dwMajorVersion == 5 && version_info.dwMinorVersion == 1 )
	windows_version = XP;
    else
	return 1; // TODO: print not supported-message

    // TODO: check if we are running on 64 bit windows
    HANDLE process = open_target_process( get_solitaire_name(windows_version), pids);
    if( !process )
    {
#ifdef DEBUG	
	fprintf(stderr,"couldn't find the '%s' process\n", get_solitaire_name(windows_version));
#else
	char buf[1024];
	sprintf_s(buf, sizeof(buf), "Couldn't find the '%s' process", get_solitaire_name(windows_version));
	MessageBox(NULL, buf, "Solitaire Trainer", MB_OK | MB_ICONERROR );
#endif
	return 1;
    }

    std::string full_name = get_full_path( get_dll_name(windows_version) );

    if( !inject_dll( process, full_name ) )
    {
#ifdef DEBUG
	fprintf(stderr,"error injecting dll\n");
#else
	MessageBox(NULL, "Error injecting dll", "Solitaire Trainer", MB_OK | MB_ICONERROR );
#endif
	CloseHandle( process );
	return 1;
    }
	    
    CloseHandle( process );
    return 0;
}
